/*
 * Copyright (C) 2008 Emweb bvba, Kessel-Lo, Belgium.
 *
 * All rights reserved.
 */

#include <boost/lexical_cast.hpp>
#include <boost/pointer_cast.hpp>

#include "Wt/WServer"
#include "WtReply.h"
#include "StockReply.h"
#include "HTTPRequest.h"
#include "WebController.h"
#include "Server.h"
#include "Utils.h"

#include <fstream>

namespace http {
  namespace server {

WtReply::WtReply(const Request& request, const Wt::EntryPoint& entryPoint,
                 const Configuration &config)
  : Reply(request),
    entryPoint_(entryPoint),
    sending_(false),
    status_(ok),
    contentLength_(-1),
    bodyReceived_(0),
    fetchMoreData_(0)
{
  if (request.contentLength > config.maxMemoryRequestSize()) {
    requestFileName_ = Wt::Utils::createTempFileName();
    // First, create the file
    std::ofstream o(requestFileName_.c_str());
    o.close();
    // Now, open it for read/write
    cin_ = new std::fstream(requestFileName_.c_str(),
      std::ios::in | std::ios::out | std::ios::binary);
    if (!*cin_) {
      // Give up, spool to memory
      requestFileName_ = "";
      delete cin_;
      cin_ = &cin_mem_;
    }
  } else {
    cin_ = &cin_mem_;
  }

  httpRequest_ = 0;
}

void WtReply::release()
{
  delete httpRequest_;
  httpRequest_ = 0;
}

WtReply::~WtReply()
{
  if (&cin_mem_ != cin_) {
    dynamic_cast<std::fstream *>(cin_)->close();
    delete cin_;
  }
  if (requestFileName_ != "") {
    unlink(requestFileName_.c_str());
  }
}

void WtReply::consumeRequestBody(Buffer::const_iterator begin,
				 Buffer::const_iterator end,
				 bool endOfRequest)
{
  if (!httpRequest_)
    httpRequest_ = new HTTPRequest(boost::dynamic_pointer_cast<WtReply>
				   (shared_from_this()), &entryPoint_);

  /*
   * Copy everything to a buffer.
   */
  if (status_ != request_entity_too_large) {
    cin_->write(begin, static_cast<std::streamsize>(end - begin));

    if (end - begin > 0) {
      bodyReceived_ += (end - begin);

      if (!connection()->server()->controller()->requestDataReceived
	  (httpRequest_, bodyReceived_, request_.contentLength)) {
	status_ = request_entity_too_large;
	setCloseConnection();
	endOfRequest = true;
      }
    }
  }

  if (endOfRequest) {
    if (status_ != ok) {
      release();
      setRelay(ReplyPtr(new StockReply(request_, status_)));
      Reply::send();
      return;
    }

    cin_->seekg(0); // rewind
    responseSent_ = false;

    HTTPRequest *r = httpRequest_;
    httpRequest_ = 0;
    connection()->server()->controller()->server_->handleRequest(r);
  }
}

void WtReply::setStatus(int status)
{
  status_ = (status_type)status;
}

void WtReply::setContentLength(::int64_t length)
{
  contentLength_ = length;
}

void WtReply::setContentType(const std::string& type)
{
  contentType_ = type;
}

void WtReply::setLocation(const std::string& location)
{
  location_ = location;
  if (status_ < 300)
    status_ = found;
}

bool WtReply::expectMoreData()
{
  return fetchMoreData_ != 0;
}

void WtReply::send(const std::string& text, CallbackFunction callBack,
		   void *cbData)
{
#ifdef WT_THREADED
  boost::recursive_mutex::scoped_lock lock(mutex_);
#endif // WT_THREADED

  fetchMoreData_ = callBack;
  callBackData_ = cbData;

  nextCout_.assign(text);

  if (!sending_) {
    sending_ = true;

    Reply::send();
  }
}

Reply::status_type WtReply::responseStatus()
{
  return status_;
}

std::string WtReply::contentType()
{
  return contentType_;
}

std::string WtReply::location()
{
  return location_;
}

::int64_t WtReply::contentLength()
{
  return contentLength_;
}

asio::const_buffer WtReply::nextContentBuffer()
{
#ifdef WT_THREADED
  boost::recursive_mutex::scoped_lock lock(mutex_);
#endif // WT_THREADED

  cout_.swap(nextCout_);

  if (!responseSent_) {
    responseSent_ = true;
    if (!cout_.empty())
      return asio::buffer(cout_);
  } else
    cout_.clear();

  while (cout_.empty() && fetchMoreData_) {
    CallbackFunction f = fetchMoreData_;
    fetchMoreData_ = 0;
    (*f)(callBackData_);
    cout_.swap(nextCout_);
  }

  // This is the last packet, unless we wait for more data
  if (cout_.empty()) {
    responseSent_ = false;
    sending_ = false;
  }

  return asio::buffer(cout_);
}

  }
}
